@@ -1,5 +1,5 @@
 
-/* Copyright (c) 2009 by Albert Graef <aggraef@gmail.com>.  Distributed
+/* Copyright (c) 2009-2017 by Albert Graef <aggraef@gmail.com>.  Distributed
    under the new BSD license, see the accompanying COPYING file for details. */
 
 #include <stdlib.h>
@@ -42,6 +42,10 @@
 #define LOG_MSGS 1
 #endif
 
+/* Global flag, determined at runtime, which indicates whether we need a main
+   signalin to make secondary signal inlets work. */
+static bool need_mainsignalin;
+
 /* These aren't in the official API, but have been around at least since Pd
    0.39, so are hopefully stable. */
 
@@ -53,7 +57,12 @@ typedef struct _namelist
 
 extern t_namelist *sys_searchpath;
 
-typedef int (*loader_t)(t_canvas *canvas, char *classname);
+/* Note that we require a version of Pd which has the loader extension API
+   (sys_register_loader et al). Also note that the most recent incarnation of
+   that API from Pd 0.47.0+ has an extra argument in the loader callback. The
+   following declarations will hopefully work with either of these, so that we
+   can adjust the interface at runtime. */
+typedef int (*loader_t)();
 extern void sys_register_loader(loader_t loader);
 
 /* Current object during creation and method call (pd_receive). */
@@ -543,10 +552,45 @@ typedef struct _runtime {
 
 static t_pure *xhead = 0, *xtail = 0;
 
+/* Helper function to map between Pd symbols and the corresponding Pure
+   names. There are some special cases here, as Pd symbols may have a
+   namespace prefix and/or a trailing tilde to indicate dsp objects. */
+
+static bool is_dsp_fun(t_symbol *sym)
+{
+  int l = strlen(sym->s_name);
+  return l>0 && sym->s_name[l-1] == '~';
+}
+
+static const char *fun_name_s(const char *name)
+{
+  /* strip off the namespace qualifier if any */
+  const char *basename = strrchr(name, '/');
+  name = basename?basename+1:name;
+  int l = strlen(name);
+  if (l>0 && name[l-1] == '~') {
+    static char buf[1024];
+    if (l >= 1020) {
+      strncpy(buf, name, 1023);
+      buf[1023] = 0;
+    } else {
+      strncpy(buf, name, l-1);
+      strcpy(buf+l-1, "_dsp");
+    }
+    return buf;
+  } else
+    return name;
+}
+
+static inline const char *fun_name(t_symbol *sym)
+{
+  return fun_name_s(sym->s_name);
+}
+
 /* We keep track of the different classes in a linked list for now. */
 
 typedef struct _classes {
-  t_symbol *sym;
+  t_symbol *sym, *fsym;
   t_class *class;
   pure_interp *interp;
   char *dir;
@@ -561,6 +605,7 @@ static void add_class(t_symbol *sym, t_class *class, char *dir)
   t_classes *new = malloc(sizeof(t_classes));
   if (!new) return;
   new->sym = sym;
+  new->fsym = gensym(fun_name(sym));
   new->class = class;
   new->interp = NULL;
   new->dir = strdup(dir);
@@ -581,7 +626,14 @@ static t_classes *lookup(t_symbol *sym)
 static t_classes *lookup_script(t_symbol *sym, const char *dir)
 {
   t_classes *act = pure_classes;
-  while (act && (act->sym != sym || (!act->interp && strcmp(act->dir, dir))))
+  /* We look up Pure classes using the corresponding Pure function symbols
+     rather than the raw Pd symbols, because the latter may be disguised by
+     namespace prefixes. Note that we still (intentionally) treat identically
+     named Pure scripts in different directories as being different objects.
+     This works because we also keep track of the script directories in the
+     Pure class table. */
+  t_symbol *fsym = gensym(fun_name(sym));
+  while (act && (act->fsym != fsym || (!act->interp && strcmp(act->dir, dir))))
     act = act->next;
   if (act)
     return act;
@@ -617,34 +669,6 @@ static pure_expr *parse_blob(void *p)
 
 /* Helper functions to convert between Pd atoms and Pure expressions. */
 
-static bool is_dsp_fun(t_symbol *sym)
-{
-  int l = strlen(sym->s_name);
-  return l>0 && sym->s_name[l-1] == '~';
-}
-
-static const char *fun_name_s(const char *name)
-{
-  int l = strlen(name);
-  if (l>0 && name[l-1] == '~') {
-    static char buf[1024];
-    if (l >= 1020) {
-      strncpy(buf, name, 1023);
-      buf[1023] = 0;
-    } else {
-      strncpy(buf, name, l-1);
-      strcpy(buf+l-1, "_dsp");
-    }
-    return buf;
-  } else
-    return name;
-}
-
-static inline const char *fun_name(t_symbol *sym)
-{
-  return fun_name_s(sym->s_name);
-}
-
 static char *get_expr(t_symbol *sym, int argc, t_atom *argv)
 {
   t_binbuf *b;
@@ -1197,11 +1221,15 @@ static void timeout(t_pure *x)
   pure_pop_interp(x_interp);
 }
 
-/* Menu callback. Thanks to Martin Peach for pointing out on pd-dev how this
-   works. */
+/* Menu edit callback. Thanks to Martin Peach for pointing out on pd-dev how
+   this works. This is used to provide an edit callback for the source file
+   associated with a Pure external, by invoking the appropriate Tcl code. We
+   use ::pd_menucommands::menu_openfile if it is available (Pd 0.43 and later)
+   and generic Tcl code otherwise. We also detect Jonathan Wilkes' Pd-L2Ork
+   variant which uses a JavaScript call instead. */
 
 #ifndef PD_MENU_COMMANDS
-#if PD_MAJOR_VERSION > 0 || PD_MINOR_VERSION >= 43
+#if PD_MAJOR_VERSION > 0 || PD_MINOR_VERSION >= 43 || defined(PDL2ORK)
 /* This needs a fairly recent Pd version (probably 0.43 or later). */
 #define PD_MENU_COMMANDS 1
 #else
@@ -1209,11 +1237,20 @@ static void timeout(t_pure *x)
 #endif
 #endif
 
+/* nw.js support. If this is non-NULL then we're running inside Jonathan
+   Wilkes' Pd-L2Ork variant and access to the GUI uses JavaScript. */
+static void (*nw_gui_vmess)(const char *sel, char *fmt, ...) = NULL;
+
 static void pure_menu_open(t_pure *x)
 {
   if (x->open_filename) {
 #if PD_MENU_COMMANDS
-    sys_vgui("::pd_menucommands::menu_openfile {%s}\n", x->open_filename);
+    if (nw_gui_vmess) {
+      char buf[PATH_MAX];
+      snprintf(buf, PATH_MAX, "%s", x->open_filename);
+      nw_gui_vmess("open_textfile", "s", buf);
+    } else
+      sys_vgui("::pd_menucommands::menu_openfile {%s}\n", x->open_filename);
 #else
     /* An older Pd version before the GUI rewrite. Let's just fire up emacs
        instead. */
@@ -1223,12 +1260,23 @@ static void pure_menu_open(t_pure *x)
     t_classes *c = x->cls;
     if (!c) return; /* this should never happen */
     if (c->dir && *c->dir) {
+      /* If there is a non-trivial namespace prefix in the object symbol, that
+	 portion of the path will already have been tacked on to the
+	 directory, so we have to use the basename of the script here. */
+      const char *basename = strrchr(c->sym->s_name, '/');
+      basename = basename?basename+1:c->sym->s_name;
 #if PD_MENU_COMMANDS
-      sys_vgui("::pd_menucommands::menu_openfile {%s/%s.pure}\n",
-	       c->dir, c->sym->s_name);
+      if (nw_gui_vmess) {
+	char buf[PATH_MAX];
+	snprintf(buf, PATH_MAX, "%s/%s.pure",
+		 c->dir, basename);
+	nw_gui_vmess("open_textfile", "s", buf);
+      } else
+	sys_vgui("::pd_menucommands::menu_openfile {%s/%s.pure}\n",
+		 c->dir, basename);
 #else
       sys_vgui("exec -- sh -c {emacs '%s/%s.pure'} &\n",
-	       c->dir, c->sym->s_name);
+	       c->dir, basename);
 #endif
     } else
       pd_error(x, "pd-pure: %s object doesn't have a script file",
@@ -1412,8 +1460,20 @@ static t_int *pure_perform(t_int *w)
 
 static void pure_dsp(t_pure *x, t_signal **sp)
 {
-  int i, n = sp[0]->s_n;
-  t_float sr = sp[0]->s_sr;
+  /* If we have a main signalin, then the leftmost inlet is a signal inlet,
+     and the real signal inlets start at an offset of 1 into sp. Otherwise, we
+     might not have any signal inlets and outlets at all, so we check that
+     before we look at the data, and assume the defaults n = 0 and sr = 0 if
+     it isn't available. XXXFIXME: A zero default value for the sample rate
+     doesn't seem to be a great idea, but currently this value apparently
+     isn't used anyway. */
+  const int offs = need_mainsignalin;
+  int i, n = 0;
+  t_float sr = 0;
+  if (x->n_dspin + x->n_dspout > 0 && sp[0]) {
+    n = sp[0]->s_n;
+    sr = sp[0]->s_sr;
+  }
   if (x->sr != sr)
     /* The sample rate at which this dsp object is supposed to run. NOTE:
        Currently this value isn't used anywhere, maybe we'd like to pass this
@@ -1427,9 +1487,9 @@ static void pure_dsp(t_pure *x, t_signal **sp)
 #endif
   dsp_add(pure_perform, 2, x, n);
   for (i = 0; i < x->n_dspin; i++)
-    x->dspin[i] = sp[i+1]->s_vec;
+    x->dspin[i] = sp[i+offs]->s_vec;
   for (i = 0; i < x->n_dspout; i++)
-    x->dspout[i] = sp[x->n_dspin+i+1]->s_vec;
+    x->dspout[i] = sp[x->n_dspin+i+offs]->s_vec;
   /* If we haven't created it yet or the block size has changed, prepare a GSL
      matrix and the corresponding Pure expression to be passed to the object
      function. The matrix has one row for each signal inlet and the row size
@@ -1774,7 +1834,13 @@ static void class_setup(const char *name, char *dir)
     class_addmethod(class, (t_method)pure_dsp, gensym((char*)"dsp"), A_NULL);
   class_addanything(class, pure_any);
   if (is_dsp) {
-    class_addmethod(class, nullfn, &s_signal, A_NULL);
+    if (need_mainsignalin) {
+      /* Create a main signalin, which is required by older Pd versions to
+	 make the other signal inlets work. */
+      class_addmethod(class, nullfn, &s_signal, A_NULL);
+      /* The above seems to be deprecated, maybe use this instead: */
+      //class_domainsignalin(class, -1);
+    }
     class_sethelpsymbol(class, gensym("pure/pure~.pd"));
   } else
     class_sethelpsymbol(class, gensym("pure/pure.pd"));
@@ -1785,7 +1851,7 @@ static void class_setup(const char *name, char *dir)
 
 /* Loader setup, pilfered from pd-lua (claudiusmaximus@goto10.org). */
 
-void class_set_extern_dir(t_symbol *s);
+EXTERN void class_set_extern_dir(t_symbol *s);
 
 static char dirbuf[MAXPDSTRING], cmdbuf[1000];
 
@@ -1812,13 +1878,28 @@ static void pure_printmsgs(char *res)
 
 #endif
 
-static int pure_loader(t_canvas *canvas, char *name)
+static int pure_loader(t_canvas *canvas, const char *name,
+		       const char *path)
 {
-  char *ptr;
-  int fd = canvas_open(canvas, name, ".pure", dirbuf, &ptr, MAXPDSTRING, 1);
+  char *ptr, namebuf[MAXPDSTRING];
+  const char *fname = name;
+  if (path && !canvas) {
+    /* The new loader in Pd 0.47.0+ indicates a library path to be searched in
+       the additional path argument. We use that here to preload Pure objects
+       at startup time in order to support the -lib mechanism. */
+    snprintf(namebuf, MAXPDSTRING, "%s/%s", path, name);
+    fname = namebuf;
+  } else if (!canvas)
+    return 0;
+  int fd = canvas_open(canvas, fname, ".pure", dirbuf, &ptr, MAXPDSTRING, 1);
   if (fd >= 0) {
     t_symbol *sym = gensym(name);
     close(fd);
+    /* We make no attempt to canonicalize the path here, but let's at least
+       strip off a trailing "/." resulting from the "./" namespace qualifier,
+       which is a fairly common idiom. */
+    char *suffix = strrchr(dirbuf, '/');
+    if (suffix && strcmp(suffix, "/.") == 0) *suffix = 0;
     class_set_extern_dir(gensym(dirbuf));
     /* Skip the loading step if the script is already loaded. */
     if (!lookup_script(sym, dirbuf)) {
@@ -1828,7 +1909,12 @@ static int pure_loader(t_canvas *canvas, char *name)
 	 rather than 'using', so that the definitions become temporaries which
 	 can be removed and reloaded later (cf. pure_reload). */
       if (chdir(dirbuf)) {}
-      sprintf(cmdbuf, "run \"%s.pure\"", name);
+      /* If there is a non-trivial namespace prefix in the object symbol,
+	 canvas_open will already haved tacked on that portion of the path to
+	 the dirbuf, so we have to use the basename of the script here. */
+      const char *basename = strrchr(name, '/');
+      basename = basename?basename+1:name;
+      sprintf(cmdbuf, "run \"%s.pure\"", basename);
 #ifdef VERBOSE
       printf("pd-pure: compiling %s.pure\n", name);
 #endif
@@ -1841,16 +1927,28 @@ static int pure_loader(t_canvas *canvas, char *name)
     }
 #ifdef EAGER
     /* Force eager compilation. */
+    pure_start_logging();
     pure_interp_compile(interp, pure_sym(fun_name_s(name)));
+    pure_stop_logging();
+    pure_printmsgs(0);
 #endif
     /* Create the object class. */
     class_setup(name, dirbuf);
     class_set_extern_dir(&s_);
-    return lookup(sym) != 0;
+    /* When invoked during startup, give some feedback that we loaded the
+       script. */
+    if (path && !canvas)
+      post("pd-pure: loaded '%s' from %s", name, dirbuf);
+    return !!lookup(sym);
   } else
     return 0;
 }
 
+static int pure_loader_legacy(t_canvas *canvas, const char *name)
+{
+  return pure_loader(canvas, name, 0);
+}
+
 static int pure_load(t_canvas *canvas, char *name)
 {
   char *ptr;
@@ -1883,6 +1981,18 @@ static int pure_load(t_canvas *canvas, char *name)
     return 0;
 }
 
+static void pure_enable_warnings(void)
+{
+  sprintf(cmdbuf, "#! --warn");
+#ifdef VERBOSE
+  printf("pd-pure: enable warnings\n");
+#endif
+  pure_start_logging();
+  char *res = pure_evalcmd(cmdbuf);
+  pure_stop_logging();
+  pure_printmsgs(res);
+}
+
 /* Reload all loaded Pure scripts. */
 
 static void reload(t_classes *c)
@@ -1905,8 +2015,12 @@ static void reload(t_classes *c)
       pure_stop_logging();
       pure_printmsgs(res);
 #ifdef EAGER
-      if (c->class)
+      if (c->class) {
+	pure_start_logging();
 	pure_interp_compile(interp, pure_sym(fun_name(c->sym)));
+	pure_stop_logging();
+	pure_printmsgs(0);
+      }
 #endif
       class_set_extern_dir(&s_);
     }
@@ -1916,13 +2030,18 @@ static void reload(t_classes *c)
 /* Create a new interpreter instance and reinitialize all objects. This must
    be invoked in a safe context, where no evaluations are in progress. */
 
+/* The command line arguments the embedded interpreter is to be invoked with.
+   This gets modified as needed in pure_setup. */
+static char *argv[] = { PD, "-I", LIBDIR "/extra/pure/lib", NULL };
+static int argc = sizeof(argv)/sizeof(char*)-1;
+
 static void pure_restart(void)
 {
   t_pure *x;
   for (x = xhead; x; x = x->next)
     pure_refini(x);
   pure_delete_interp(interp);
-  interp = pure_create_interp(0, 0);
+  interp = pure_create_interp(argc, argv);
   pure_switch_interp(interp);
 #ifdef VERBOSE
   printf("pd-pure: restarting, please wait...\n");
@@ -1982,7 +2101,9 @@ static void *runtime_init(t_symbol *s, int argc, t_atom *argv)
   x->out2 = outlet_new(&x->obj, 0);
   for (i = 0; i < argc; i++)
     if (argv[i].a_type != A_SYMBOL)
-      pd_error(x, "pure-runtime: bad argument #%d, expected filename", i+1);
+      pd_error(x, "pure-runtime: bad argument #%d, expected filename or -w", i+1);
+    else if (!strcmp(argv[i].a_w.w_symbol->s_name, "-w"))
+      pure_enable_warnings();
     else if (!pure_load(canvas, argv[i].a_w.w_symbol->s_name))
       pd_error(x, "pure-runtime: error loading script '%s.pure'",
 	       argv[i].a_w.w_symbol->s_name);
@@ -2040,21 +2161,58 @@ extern int pure_register_class(const char *name, pure_interp *interp,
 
 /* Loader setup. */
 
+#ifdef WIN32
+#include <windows.h>
+#else
+#define __USE_GNU // to get RTLD_DEFAULT
+#include <dlfcn.h> // for dlsym
+#ifndef RTLD_DEFAULT
+/* If RTLD_DEFAULT still isn't defined then just passing NULL will hopefully
+   do the trick. */
+#define RTLD_DEFAULT NULL
+#endif
+#endif
+
 extern void pure_setup(void)
 {
-  char buf[MAXPDSTRING];
+  static char buf[MAXPDSTRING];
   char *ptr;
   int fd;
-  interp = pure_create_interp(0, 0);
+  int major, minor, bugfix;
+  sys_getversion(&major, &minor, &bugfix);
+  /* Check whether we need to work around a bug in older Pd versions which
+     breaks secondary signal inlets if no main signalin is present. This bug
+     has been fixed in 0.47.0, so we require the workaround for all minor
+     versions before that. Note that this has the side-effect of turning the
+     leftmost inlet into a dummy signal inlet, which is confusing for the
+     user, but there's no way around that for Pd 0.46 and earlier. */
+  need_mainsignalin = minor < 47;
+  /* Determine the pd-pure include path at runtime. */
+  strncpy(buf, get_libdir("/extra/pure/lib"), MAXPDSTRING);
+  argv[2] = buf;
+  interp = pure_create_interp(argc, argv);
   if (interp) {
     pure_expr *x = pure_symbol(pure_sym("version"));
     char *pure_version = 0;
     pure_is_cstring_dup(x, &pure_version);
-    post("pd-pure %s (pure-%s) (c) 2009-2014 Albert Graef <aggraef@gmail.com>", VERSION, pure_version);
-    post("pd-pure: compiled for %s-%d.%d on %s %s", PD, PD_MAJOR_VERSION, PD_MINOR_VERSION, __DATE__, __TIME__);
+    post("pd-pure %s (pure-%s) (c) 2009-2017 Albert Graef <aggraef@gmail.com>", VERSION, pure_version);
+#ifdef PD_L2ORK_VERSION
+    /* The latest Purr Data (a.k.a. Pd-l2ork 2.x) versions have a proper
+       version number now, which is different from the vanilla compatibility
+       version, so use that instead of the generic flavor-vanilla-version
+       moniker. We still indicate the vanilla compatibility version in
+       parentheses, however, since this determines the revision of the Pd API
+       that the source was compiled against. */
+    post("pd-pure: compiled for Pd-l2ork-%s (Pd %d.%d) on %s", PD_L2ORK_VERSION, PD_MAJOR_VERSION, PD_MINOR_VERSION, __DATE__);
+#else
+    post("pd-pure: compiled for %s-%d.%d on %s", PD, PD_MAJOR_VERSION, PD_MINOR_VERSION, __DATE__);
+#endif
     if (pure_version) free (pure_version);
     /* Register the loader for Pure externals. */
-    sys_register_loader(pure_loader);
+    if (major > 0 || minor >= 47)
+      sys_register_loader(pure_loader);
+    else
+      sys_register_loader(pure_loader_legacy);
     /* Create the proxy class for extra inlets. */
     px_class = class_new(gensym("pure proxy"), 0, 0,
 			 sizeof(t_px), CLASS_PD|CLASS_NOINLET,
@@ -2078,6 +2236,19 @@ extern void pure_setup(void)
     save_sym = pure_sym("pd_save");
     restore_sym = pure_sym("pd_restore");
     varargs_sym = pure_sym("varargs");
+#if PD_MENU_COMMANDS
+    /* nw.js support. */
+#ifdef WIN32
+    nw_gui_vmess = (void*)GetProcAddress(GetModuleHandle("pd.dll"), "gui_vmess");
+#else
+    nw_gui_vmess = dlsym(RTLD_DEFAULT, "gui_vmess");
+#endif
+    if (nw_gui_vmess)
+      post("pd-pure: using JavaScript interface (Pd-l2ork nw.js version)");
+#endif
+#ifdef VERBOSE
+    post("pd-pure: adding '%s' to include path", buf);
+#endif
   } else
     error("pd-pure: error initializing interpreter; loader not registered");
 }
